from datetime import datetime

import pydantic
from envoy_schema.admin.schema import base


class AggregatorRequest(pydantic.BaseModel):
    """Basic attributes for the creation of a new aggregator."""

    name: str


class AggregatorDomain(pydantic.BaseModel):
    """A domain whitelisted by the utility server as being controlled by the parent aggregator"""

    domain: str
    created_time: datetime
    changed_time: datetime


class AggregatorResponse(pydantic.BaseModel):
    """Metadata about a specific aggregator registered in the utility server"""

    aggregator_id: int
    name: str
    domains: list[AggregatorDomain]


class AggregatorPageResponse(base.BasePageModel):
    """Represents a paginated response of Aggregators"""

    aggregators: list[AggregatorResponse]  # The aggregator models in this page


class AggregatorDomainResponse(pydantic.BaseModel):
    """Represents a specific aggregator domain registered in the utility server"""

    aggregator_domain_id: int
    aggregator_id: int
    created_time: datetime
    changed_time: datetime
    domain: str


class AggregatorDomainPageResponse(base.BasePageModel):
    """Represents a paginated response of aggregator domains"""

    aggregator_domains: list[AggregatorDomainResponse]



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel


class SiteControlRequest(BaseModel):
    """Used for encoding a "SiteControl" which can represent things like a Dynamic Operating Envelope, Setpoint or
    other form of control"""

    site_id: int  # Corresponds to EndDevice id - the site that this control is targeting
    calculation_log_id: Optional[int]  # The ID of the CalculationLog that created this control (or NULL if no link)
    duration_seconds: int
    start_time: datetime

    randomize_start_seconds: Optional[int] = (
        None  # A number of seconds from -3600 to 3600 that a site should treat as a random range to vary start_time by
    )

    set_energized: Optional[bool] = None  # Corresponds to CSIP-Aus opModEnergize (None will not encode anything)
    set_connect: Optional[bool] = None  # Corresponds to CSIP-Aus opModConnect (None will not encode anything)

    import_limit_watts: Optional[Decimal] = None  # Corresponds to CSIP-Aus opModImpLimW (None will not encode anything)
    export_limit_watts: Optional[Decimal] = None  # Corresponds to CSIP-Aus opModExpLimW (None will not encode anything)
    generation_limit_watts: Optional[Decimal] = (
        None  # Corresponds to CSIP-Aus opModGenLimW (None will not encode anything)
    )
    load_limit_watts: Optional[Decimal] = None  # Corresponds to CSIP-Aus opModLoadLimW (None will not encode anything)
    set_point_percentage: Optional[Decimal] = (
        None  # percent of device max power settings to charge (if negative) or discharge (if positive) at. 100 = 100%
    )
    ramp_time_seconds: Optional[Decimal] = (
        None  # Corresponds to rampTms (None will not encode anything). 100 = 100 seconds
    )


class SiteControlResponse(SiteControlRequest):
    """Site Control basic model when being queried externally"""

    site_control_id: int  # Internal identifier for this DOE
    created_time: datetime  # When this control was created
    changed_time: datetime  # When this control was last changed
    superseded: bool  # Whether this control has been superseded by a higher priority control since its creation


class SiteControlPageResponse(BaseModel):
    """Represents a paginated response of SiteControlResponse"""

    total_count: int  # The total number of controls (independent of this page of results)
    limit: int  # The maximum number of controls that could've been returned (the limit set by the query)
    start: int  # The number of controls that have been skipped as part of this query (the start set by the query)
    after: Optional[datetime]  # The "after" filter set by the query
    controls: list[SiteControlResponse]  # The control models in this paged response


class SiteControlGroupRequest(BaseModel):
    """Used for creating new SiteControlGroups (used for grouping SiteControls)"""

    description: str  # Human readable description (32 char max)
    primacy: int  # Lower = Higher priority. Affects "child" controls relative priority when compared to other groups
    fsa_id: int = 1  # The function set assignment ID that this SiteControl group will be grouped under


class SiteControlGroupResponse(SiteControlGroupRequest):
    """Represents a server side view of current SiteControls"""

    site_control_group_id: int  # Primary key
    created_time: datetime  # When this group was created
    changed_time: datetime  # When this group was last changed


class SiteControlGroupPageResponse(BaseModel):
    """Represents a page of SiteControlGroups"""

    total_count: int  # The total number of groups (independent of this page of results)
    limit: int  # The maximum number of groups that could've been returned (the limit set by the query)
    start: int  # The number of groups that have been skipped as part of this query (the start set by the query)
    after: Optional[datetime]  # The "after" filter set by the query
    site_control_groups: list[SiteControlGroupResponse]  # The control group models in this paged response



from datetime import datetime
from decimal import Decimal
from enum import IntEnum

from pydantic import BaseModel


class PhaseEnum(IntEnum):
    """Condensed csip-aus compliant phase enum of the sep2 types."""

    NA = 0
    AN = 129
    BN = 65
    CN = 33


class CSIPAusSiteReadingUnit(IntEnum):
    """
    This enum aims to capture allowed csip aus compliant readings for a site as a single value.

    There is some flexibility built in for readings which aren't perfectly compliant (e.g. Kind or data qualifier of 0).
    A future release might include a strict bool flag to enable the use of imperfect readings.

    It assumes the following:
        - The Kind is 37 (POWER, mandatory by csip) or 0 (NA)
        - The data qualifier is 2 (Average, mandatory by csip) or 0 (NA)
        - Any RoleFlags are allowed, does not show whether sites are mirrors, DER, submeter, or aggregation points.
            If both site and device information are present (role flags 3 and 73, respectively, site data is returned)
        - Load convention applies: positive is import from the grid. If FlowDirectionType is stored it will be used,
            otherwise it is provided as stored (e.g. Flow type = 0 is not converted).
        - Other options are not filtered (DeviceCategory, CommodityType, etc). All are allowed, none are returned.
    """

    ACTIVEPOWER = 1  # W, UOM =  38
    REACTIVEPOWER = 2  # Var, UOM = 63
    FREQUENCY = 3  # Hz, UOM = 33
    VOLTAGE = 4  # V, UOM = 29


class CSIPAusSiteReading(BaseModel):
    """
    Admin representation of a csip aus compliant reading for a specific site and reading unit (CSIPReadingUnit).
    """

    reading_start_time: datetime
    duration_seconds: int
    phase: PhaseEnum
    value: Decimal
    csip_aus_unit: CSIPAusSiteReadingUnit  # # CSIP Aus unit enum used for the query


class CSIPAusSiteReadingPageResponse(BaseModel):
    """Paginated response for csip aus site reading queries."""

    total_count: int  # Total number of readings matching the query
    limit: int  # Maximum number of readings that could be returned
    start: int  # Number of readings skipped for pagination
    site_id: int  # Site ID filter used in the query
    start_time: datetime  # Start time filter used in the query
    end_time: datetime  # End time filter used in the query
    readings: list[CSIPAusSiteReading]  # The readings in this page



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel


class UpdateDefaultValue(BaseModel):
    """Used to allow the setting of None or a specific value"""

    value: Optional[Decimal]  # The decimal value to update (or None to update the value to None)


class ControlDefaultRequest(BaseModel):
    """Used for updating the "failover" SiteControl defaults that are used if no specific control is active"""

    import_limit_watts: Optional[
        UpdateDefaultValue
    ]  # If set - update The default "import limit watts" used in absence of an active control
    export_limit_watts: Optional[
        UpdateDefaultValue
    ]  # If set - update The default "export limit watts" used in absence of an active control
    generation_limit_watts: Optional[
        UpdateDefaultValue
    ]  # If set - update The default "generation limit watts" used in absence of an active control
    load_limit_watts: Optional[
        UpdateDefaultValue
    ]  # If set - update The default "export limit watts" used in absence of an active control
    ramp_rate_percent_per_second: Optional[
        UpdateDefaultValue
    ]  # If set - update The default ramp rate expressed in a percent of max W per second


class ControlDefaultResponse(BaseModel):
    """Snapshot of the current "failover" SiteControl defaults that are used if no specific control is active"""

    server_default_import_limit_watts: Optional[Decimal]  # None means NO default
    server_default_export_limit_watts: Optional[Decimal]
    server_default_generation_limit_watts: Optional[Decimal]
    server_default_load_limit_watts: Optional[Decimal]
    ramp_rate_percent_per_second: Optional[Decimal]

    created_time: datetime
    changed_time: datetime


class RuntimeServerConfigRequest(BaseModel):
    """Used for updating the server configuration associated with runtime behavior"""

    dcap_pollrate_seconds: Optional[int] = None  # If set - update the DeviceCapability pollRate
    edevl_pollrate_seconds: Optional[int] = None  # If set - update the EndDeviceList pollRate
    fsal_pollrate_seconds: Optional[int] = None  # If set - update the FunctionSetAssignmentList pollRate
    derpl_pollrate_seconds: Optional[int] = None  # If set - update the DERProgram pollRate
    derl_pollrate_seconds: Optional[int] = None  # If set - update the DERList pollRate
    mup_postrate_seconds: Optional[int] = None  # If set - update the MirrorUsagePoint postRate

    # pow10 Values affect the associated integer sent. Eg for a "watts" value, setting pow10 to -3 will mean that
    # the encoded value of 12345 should be interpreted as 12.345 watts by the client. Setting a value of 2 will mean
    # that a value of 12345 should be interpreted as 1234500 watts.
    site_control_pow10_encoding: Optional[int] = None  # If set - update the pow10 encoded for sent DERControl values.
    tariff_pow10_encoding: Optional[int] = None  # If set - update the pow10 encoded for sent RateComponent values

    disable_edev_registration: Optional[bool] = None  # If True - EndDevice RegistrationLink's will be disabled


class RuntimeServerConfigResponse(BaseModel):
    """Snapshot of the server configuration associated with runtime behavior"""

    dcap_pollrate_seconds: int
    edevl_pollrate_seconds: int
    fsal_pollrate_seconds: int
    derpl_pollrate_seconds: int
    derl_pollrate_seconds: int
    mup_postrate_seconds: int

    site_control_pow10_encoding: int
    tariff_pow10_encoding: int

    disable_edev_registration: bool

    created_time: datetime
    changed_time: datetime



import pydantic
import datetime as dt

from . import base


class CertificateResponse(pydantic.BaseModel):
    """Represents a specific certificate registered in the utility server"""

    certificate_id: int
    created: dt.datetime
    lfdi: str
    expiry: dt.datetime


class CertificatePageResponse(base.BasePageModel):
    """Represents a paginated response of certificates"""

    certificates: list[CertificateResponse]


class CertificateRequest(pydantic.BaseModel):
    """Represents a specific certificate to be created by the utility server"""

    lfdi: str
    expiry: dt.datetime


class CertificateAssignmentRequest(pydantic.BaseModel):
    """Represents a single assignment of a certificate to another entity

    The intent is that the certificate can be created from the model or simply assigned e.g. to an aggregator
    """

    certificate_id: int | None = None
    lfdi: str | None = None
    expiry: dt.datetime | None = None

    @pydantic.model_validator(mode="after")
    def id_or_lfdi_provided_xor(self) -> "CertificateAssignmentRequest":
        """Validates either one of lfdi or id provided, but not both (exclusive or).

        Both are identifiers of a single certificate, so ambiguity could be
        created if both are provided. To avoid, this model rejects validation.

        Raises:
            ValueError if neither id or lfdi provided, or if both are provided
        """
        if self.certificate_id is None and self.lfdi is None:
            raise ValueError("Either lfdi or id needs to be provided")
        elif self.certificate_id is not None and self.lfdi is not None:
            raise ValueError("Only one of lfdi or id can be provided, not both")
        return self
